@@ -40,10 +40,10 @@ static int ssdp_parse_lines(struct ndpi_detection_module_struct *ndpi_struct,
     packet->packet_lines_parsed_complete = 0;
     ndpi_parse_packet_line_info_any(ndpi_struct, flow);
 #if 0
-     printf("SSDP: parsed_completed: %d, parsed_lines: %d\n",
-        packet->packet_lines_parsed_complete,
-        packet->parsed_lines
-    );
+    printf("SSDP: parsed_completed: %d, parsed_lines: %d\n",
+            packet->packet_lines_parsed_complete,
+            packet->parsed_lines
+          );
 
     for (int i = 0; i < packet->parsed_lines; i++) {
         if (packet->line[i].len <= 2) continue;
@@ -57,28 +57,11 @@ static int ssdp_parse_lines(struct ndpi_detection_module_struct *ndpi_struct,
     return packet->parsed_lines;
 }
 
-int ssdp_extra_processing(struct ndpi_detection_module_struct *ndpi_struct,
-        struct ndpi_flow_struct *flow) {
-
-    struct ndpi_packet_struct *packet = &flow->packet;
-
-    ssdp_parse_lines(ndpi_struct, flow);
-
-    if (flow->max_extra_packets_to_check > 1)
-        return 0;
-
-    return 1;
-}
-
 static void ndpi_int_ssdp_add_connection(struct ndpi_detection_module_struct
         *ndpi_struct, struct ndpi_flow_struct *flow)
 {
     struct ndpi_packet_struct *packet = &flow->packet;
 
-    flow->check_extra_packets = 1;
-    flow->max_extra_packets_to_check = 2;
-    flow->extra_packets_func = ssdp_extra_processing;
-
     ssdp_parse_lines(ndpi_struct, flow);
 
     ndpi_set_detected_protocol(ndpi_struct, flow, NDPI_PROTOCOL_SSDP);
@@ -94,17 +77,17 @@ void ndpi_search_ssdp(struct ndpi_detection_module_struct *ndpi_struct, struct n
     if (packet->udp != NULL) {
 
         if (packet->payload_packet_len >= SSDP_MSEARCH_LEN &&
-            memcmp(packet->payload, SSDP_MSEARCH, SSDP_MSEARCH_LEN) == 0) {
-                NDPI_LOG_INFO(ndpi_struct, "found ssdp (M-SEARCH)\n");
-                ndpi_int_ssdp_add_connection(ndpi_struct, flow);
-                return;
+                memcmp(packet->payload, SSDP_MSEARCH, SSDP_MSEARCH_LEN) == 0) {
+            NDPI_LOG_INFO(ndpi_struct, "found ssdp (M-SEARCH)\n");
+            ndpi_int_ssdp_add_connection(ndpi_struct, flow);
+            return;
         }
 
         if (packet->payload_packet_len >= SSDP_NOTIFY_LEN &&
-            memcmp(packet->payload, SSDP_NOTIFY, SSDP_NOTIFY_LEN) == 0) {
-                NDPI_LOG_INFO(ndpi_struct, "found ssdp (NOTIFY)\n");
-                ndpi_int_ssdp_add_connection(ndpi_struct, flow);
-                return;
+                memcmp(packet->payload, SSDP_NOTIFY, SSDP_NOTIFY_LEN) == 0) {
+            NDPI_LOG_INFO(ndpi_struct, "found ssdp (NOTIFY)\n");
+            ndpi_int_ssdp_add_connection(ndpi_struct, flow);
+            return;
         }
     }
 
